#include <time.h>
#include <sys/sysctl.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include "general.h"





/*************/
/* FUNCTIONS */
/*************/

/* Append the list of labels to the sens-array of the current resource
   Dynamically resizes this array if necessary  */
void appendToSens(int curr, const int* labels, int num_of_labels, Marks *marks) {
    marks[curr].sens.insert(marks[curr].sens.end(), &labels[0], &labels[num_of_labels]);
}
/* Append the list of labels to the imm-array of the current resource
   Dynamically resizes this array if necessary  */
void appendToImm(int curr, const int* labels, int num_of_labels, Marks *marks) {
    marks[curr].imm.insert(marks[curr].imm.end(), &labels[0], &labels[num_of_labels]);
}
/* Append the list of labels leading to 2marked neighbors to the imm-array of the current resource
   Dynamically resizes this array if necessary  */
int appendToImm_2marks(int curr, Graph *g, Marks *marks) {
    int i;
    int ret = 0;
    int num_of_edges = g->n;
    for (i=0; i<num_of_edges; i++) {
        if (marks[g->neighbor(i)].marking == 2 && g->labelsAddedToImm(i) == 0) {
            appendToImm(curr, g->labels(i), g->labelCount(i), marks);
            g->setLabelsAddedToImm(i, 1);
            ret = 1;
        }
    }
    return ret;
}



/* returns 1 iff there is an element a \in x such that a \in y and a \notin z */
//int exists_shared_member_not_in(int *x, int size_x, int *y, int size_y,int *z, int size_z) {
int exists_shared_member_not_in(const int *x, int size_x, const Marks::Container &y, const Marks::Container &z) {
    for (int i = size_x-1; i >= 0; --i) {
        int xv = x[i];
        for(int j = y.size()-1; j >= 0; --j)
            if (xv == y[j]) {
                bool found = false;
                for(int k=z.size()-1; k >= 0 && !found; --k)
                    found = xv == z[k];
                if (!found)
                    return true;
            }
    }
    return false;
}

/* returns 1 iff there is an element a \in x such that a \in y */
int exists_shared_member(const int *x, int size_x, const int *y, int size_y) {
    for (int i = size_x-1; i >= 0; --i) {
        int xv = x[i];
        for(int j = size_y-1; j >= 0; --j)
            if (xv == y[j])
                return true;
    }
    return false;
}
/* creates a copy of the list */
struct int_List cpy_to_int_List(const int *ints, int n) {
    int i = 0;
    int *ret_elts = (int*) malloc (n * sizeof(int));
    for (i=0; i<n; i++)
        ret_elts[i] = ints[i];
    struct int_List ret = {n,ret_elts};
    return ret;
}
/* returns an int_List struct, with the intersection of the elements in `il' and the elements of `x'.
   Assumes no duplicate elements in both lists
*/
struct int_List intersect(int *y, int size_of_y, int *x, int size_of_x) {
    int *ret = new int[size_of_x];
    int i,j;
    int n = 0;
    for(i=size_of_y-1; i >= 0; --i) {
        int yv = y[i];
        for (j=size_of_x-1; j >= 0; --j) {
            if (yv == x[j]) {
                ret[n++] = yv;
                break;
            }
        }
    }
    struct int_List r = {n,ret};
    return r;
}
/* Append the list of labels to the imm-array of the current resource
   Does not add duplicate elements
   Dynamically resizes this array if necessary  */
void append_to_dyn_int_List(struct dyn_int_List &x, int* y, int num_of_y) {
    if(x.num_elts + num_of_y >= x.num_allocated) {					// Are more refs required?
        if (x.num_allocated == 0) {
            x.num_allocated = INITIAL_NUM_OF_LABELS;
            x.num_allocated += num_of_y;
        }
        else {														// Expand list, each time at least doubling the size
            x.num_allocated += num_of_y;
            x.num_allocated *= 2;
        }
        void *_tmp = realloc(x.elts, (x.num_allocated * sizeof(int)));
        if (_tmp == NULL)
        {
            fprintf(stderr, "ERROR: Couldn't realloc memory!\n");
            puts("");
            return;
        }
        x.elts = (int*)_tmp;
    }

    int i,j,found;
    for (i=0; i<num_of_y; i++) {
		found = 0;
    	for (j=0; j<x.num_elts&&found==0; j++)
			found = x.elts[j] == y[i] ? 1 : 0;
		if (found == 0)
			x.elts[x.num_elts++] = y[i];
    }
}


/* Returns t iff there is a path \pi that satisfies the following properties:
   \pi starts in `curr'
   \pi is a d-path for some label d
   the last element of \pi (the head) is a 3-marked resource where label d is included in the sens-array
   */
//TODO: it has not mechanically been proven that this function satisfies its specification
int E_d_path_to_013_marked_resource (Graph *g, int curr, int curr_index, int prev, struct int_List dests, int first_call, Marks *marks, struct dyn_int_List *dpath_dests, int* visited_013) {
    int i;
    int ret = 0;

    if (dests.n == 0) {
    }
    else if (marks[curr].marking == 0 || marks[curr].marking == 1 || (marks[curr].marking == 3 &&
             (first_call == 1 || exists_shared_member_not_in(dests.elts,dests.n,marks[curr].sens,marks[curr].imm)))) {
        ret = 1;
    }
    else if (visited_013[curr] == 1) {
		ret = marks[curr].marking == 3 && exists_shared_member(dests.elts,dests.n,dpath_dests[curr].elts,dpath_dests[curr].num_elts) ? 1 : 0;
        //ret = exists_shared_member(dests.elts,dests.n,dpath_dests[curr].elts,dpath_dests[curr].num_elts) ? 1 : 0;
    }
    else if (marks[curr].marking != 2) {
        visited_013[curr] = 1;
        struct int_List new_dests;
        for (i=0; i<g[curr].n && ret!=1; i++) {
            if (first_call == 1)
                new_dests = cpy_to_int_List(g[curr].labels(i), g[curr].labelCount(i));
            else
                new_dests = intersect(dests.elts, dests.n, g[curr].labels(i), g[curr].labelCount(i));
            ret = E_d_path_to_013_marked_resource(g,g[curr].neighbor(i),i,curr,new_dests,0,marks,dpath_dests,visited_013);
            delete[] new_dests.elts;
        }

        if (marks[curr].marking == 3 && prev != -1) {
            append_to_dyn_int_List(dpath_dests[prev], g[prev].labels(curr_index), g[prev].labelCount(curr_index));
        }
        else if (marks[curr].marking == 4 && prev != -1) {
            struct int_List new_dpath_dests = intersect(dpath_dests[curr].elts, dpath_dests[curr].num_elts, g[prev].labels(curr_index), g[prev].labelCount(curr_index));
            append_to_dyn_int_List(dpath_dests[prev], new_dpath_dests.elts, new_dpath_dests.n);
            delete[] new_dpath_dests.elts;
        }
        visited_013[curr] = 0;
    }
    return ret;
}

// A wrapper to initialize data structures for a call to function `E_d_path_to_013_marked_resource'.
int E_d_path_to_013 (Graph *g, int curr, Marks *marks, int* visited_013, int num_of_resources) {
    int ret;
    struct int_List dests = {-1,NULL};
    struct dyn_int_List *dpath_dests = new dyn_int_List[num_of_resources];
    init_dyn_int_List(dpath_dests, num_of_resources);
zero_list(visited_013, num_of_resources);
    ret = E_d_path_to_013_marked_resource(g,curr,-1,-1,dests,1,marks,dpath_dests,visited_013);
    for (int i = 0; i < num_of_resources; ++i)
        if (dpath_dests[i].elts)
            delete dpath_dests[i].elts;
    delete[] dpath_dests;
    // TODO: moet dests niet ook gefreed worden?
    return ret;
}








/*****************/
/* THE ALGORITHM */
/*****************/


void step1 (Graph *g, int curr, int curr_index, int parent, Marks *marks, int *nodeset, int num_of_resources, int* visited_013, const Params *params) {
    int i;

    if (curr < 0 || curr >= num_of_resources) fprintf(stderr, "ERROR: Current resource illegal: %i\n",curr);
    if (nodeset[curr] == 0) fprintf(stderr, "ERROR: current node illegal %i (%i)\n", curr, parent);
    if (parent != -1 && (curr_index < 0 || curr_index >= g[parent].n)) fprintf(stderr, "ERROR: current index illegal: %i,%i,%i\n",curr,curr_index,g[curr].n);


    /* If the current resource is marked 1,3 or 4, the parent must have another way to escape the destinations leading from parent to curr.
       Thus we add these destinations to the sens array of the parent resource */
    if (marks[curr].marking == 1 || marks[curr].marking == 3 || marks[curr].marking == 4) {
        if (parent != -1)
            appendToSens(parent, g[parent].labels(curr_index), g[parent].labelCount(curr_index), marks);
    }
    /* If the current resource is marked 2 or is a sink, then the parent has an escape for the destinations leading from parent to curr.
       Thus we add these destinations to the imm array of the parent resource */
    else if (marks[curr].marking == 2 || g[curr].n == 0) {
        marks[curr].marking = 2;
        if (parent != -1)
            appendToImm(parent, g[parent].labels(curr_index), g[parent].labelCount(curr_index), marks);
    }
    /* Otherwise: mark the current resource 1 because it is visited. Recursively perform a forward search, during which information is propagated backwards
       to the current resource. */
    else {
        marks[curr].marking = 1;
        for (i=0; i < g[curr].n; i++) {
            step1(g, g[curr].neighbor(i), i, curr, marks,nodeset,num_of_resources,visited_013,params);
        }
        /* If the current resource has a destination in sens but not in imm, mark it 3 and propagate information to the parent */
        if (!(marks[curr].subset())) {
            marks[curr].marking = 3;
            if (parent != -1)
                appendToSens(parent, g[parent].labels(curr_index), g[parent].labelCount(curr_index), marks);
        }
        /* If there is a path to a 0,1 or 3 marked resource, mark it 4 and propagate information backward */
        else {
            marks[curr].marking = 4;
            //if (E_d_path_to_013(g, curr, marks, visited_013, num_of_resources) == 1) {
            if (marks[curr].sens.size() > 0) {
                if (parent != -1)
                    appendToSens(parent, g[parent].labels(curr_index), g[parent].labelCount(curr_index), marks);
            }
            /* else mark the resource 2 and propagate information backward */
            else {
                marks[curr].marking = 2;
                if (parent != -1) {
                    appendToImm(parent, g[parent].labels(curr_index), g[parent].labelCount(curr_index), marks);
                    g[parent].setLabelsAddedToImm(curr_index, 1);
                }
            }
        }
    }
}

bool step2 (Graph *g, int curr, Marks *marks, int *nodeset, int num_of_resources, int* visited_013, const Params *params, int flag);
void step2_for_34parents(Graph *g, int curr, Marks *marks, int *nodeset, int num_of_resources, int* visited_013, const Params *params, int flag) {
    for (int j=0; j<g[curr].parents_elements; j++)
        if (marks[g[curr].parents[j]].marking==3 || marks[g[curr].parents[j]].marking == 4)
            step2(g, g[curr].parents[j], marks,nodeset,num_of_resources, visited_013, params, flag);
}
bool step2 (Graph *g, int curr, Marks *marks, int *nodeset, int num_of_resources, int* visited_013, const Params *params, int flag) {
    int n;
    bool ret = false;

    if (marks[curr].marking == 3) {
        n = appendToImm_2marks(curr, &g[curr], marks);
        if (n == 1) {
            if (marks[curr].subset()) {
                ret = true;
                marks[curr].marking = 4;
                if (E_d_path_to_013(g, curr, marks, visited_013, num_of_resources) == 1)
                    step2_for_34parents(g,curr,marks,nodeset,num_of_resources,visited_013, params, flag);
                else {
                    marks[curr].marking = 2;
                    step2_for_34parents(g,curr,marks,nodeset,num_of_resources,visited_013, params, flag);
                }
            }
        }
    }
    else if (marks[curr].marking == 4) {
        if (E_d_path_to_013(g, curr, marks, visited_013, num_of_resources) != 1) {
            ret = true;
            n = appendToImm_2marks(curr, &g[curr], marks);
            marks[curr].marking = 2;
            step2_for_34parents(g,curr,marks,nodeset,num_of_resources,visited_013, params, flag);
        }
    }
    return ret;
}













int WHS_run_algorithm_on_faulty_configuration(ResourceList* faulty_resources, const Params* params) {
	if (output)
    		std::cout << "Starting algorithm for configuration with faulty channels: " << *faulty_resources << std::endl;

    // Declare variables
    int num_of_resources;
    int i,j,dl,parent,curr_index,printed_suppress_warning, found_4_mark;
	bool not_at_fixpoint;

    // Initialize assisting parameter, determine number of resources, reserve space for data structures
    num_of_resources = bitsize_of_resources(params);
    Graph *g = new Graph[num_of_resources];
    Marks *marks = new Marks[num_of_resources];
    int *nodeset = new int[num_of_resources];
    int* visited_013 = new int[num_of_resources];

    // Initialize graph and markings data structures
    zero_list(nodeset,num_of_resources);
    dl = fill_graph(g,nodeset,faulty_resources,params);
	if (dl == RET_DISC_ROUTING)
		goto TERMINATE_RUN_ALGORITHM;
    compute_parents(g, nodeset, num_of_resources);
    zero_list(visited_013,num_of_resources);


    // Step 1
    for (i=0; i<num_of_resources; i++) {
        if (nodeset[i] == 1) {
            curr_index = -1;
            parent = -1;
            step1(g, i, curr_index, parent, marks,nodeset,num_of_resources,visited_013,params);
        }
    }

    // Step 2
    not_at_fixpoint = true;
    while (not_at_fixpoint) {
        not_at_fixpoint = false;
        for (i=0; i<num_of_resources; i++) {
            if (nodeset[i] == 1) {
    			zero_list(visited_013,num_of_resources);
                not_at_fixpoint = step2(g,i,marks,nodeset,num_of_resources,visited_013, params,0) || not_at_fixpoint;
			}
        }
    }

    // Find a 3/4-marked resource
    dl = RET_DLF, printed_suppress_warning = 0, found_4_mark = 0;
    for (i=0; i<num_of_resources; i++) {
        if ((marks[i].marking == 3 || marks[i].marking == 4)) {
			found_4_mark |= marks[i].marking == 4;
           	if (dl == RET_DLF) {
           	    dl = RET_DL_NOT_MIN;
           	    printf("Found possible deadlock under fault model:\n");
           	  	int j= 0;
           	    while (faulty_resources->at(j).valid()) {
           	        std::cout << faulty_resources->at(j++);
           	    	printf(", ");
            	}
				printf("\n");
			}
			if (inst_convert_to_resource(i,params).p != L) {
       	     	int j, printed=0;
            	std::cout << "Resource " << inst_convert_to_resource(i,params);
				if (marks[i].marking == 3)
            		printf(" can possibly be in deadlock if it contains a header/tail-flit with a destination in {");
				else
        	    	printf(" can possibly be in deadlock if it contains a tail-flit with a destination in {");
        	    for (j=0; j<marks[i].sens.size()&&printed<MAX_DESTS_PER_RESOURCE; j++) {
           	    	int jj,found=0;
           	    	for (jj=0; jj<marks[i].imm.size()&&found==0; jj++)
                    	found = (marks[i].imm[jj] == marks[i].sens[j] ? 1 : 0);
                	if (found == 0 || marks[i].marking == 4) {
                   		inst_print_procnode (marks[i].sens[j], params);
                    	printf(",");
                    	printed++;
                	}
           		}
            	if (printed == MAX_DESTS_PER_RESOURCE)
                	printf("...");
           		printf("}\n");
			}
			else if (printed_suppress_warning == 0) {
				// Suppresssing printing deadlocks in local channels
				printed_suppress_warning = 1;
				printf("Suppressing printing output on local channels.\n");
			}
        }
        else if (marks[i].marking != 2 && nodeset[i] == 1) {
            std::cout << "ERROR: resource " << inst_convert_to_resource(i,params);
            printf(" is marked %i:\n",marks[i].marking);
        }
    }
	if (dl == RET_DL_NOT_MIN && (found_4_mark || !MIN_DL_NOT_NEC)) {
		if (!DO_NOT_CHK_R_OF_POS_DL) {
			if (output)
				printf("Starting to check reachability of possible deadlock...\n");
			dl = check_reachability_possible_deadlock(g, marks, params, faulty_resources);
			switch (dl) {
			 case RET_DLF:
				if (output)
					printf("Ruled out unreachable deadlocks. The network is deadlock-free.\n");
				break;
			 case RET_DL_MIN:
				if (output)
					printf("Found reachable deadlock. The printed deadlock is minimal.\n");
				break;
			 case RET_DL_NOT_MIN:
				if (output)
					printf("Found reachable deadlock. The printed deadlock is not necessarily minimal.\n");
				break;
			 default:
				if (output)
					printf("The user aborted the solver or an error occured while determining legallity of possible deadlock.\n");
			}
		}
	}
	else if (dl == RET_DL_NOT_MIN && !found_4_mark) {
		if (output)
			printf("Possible deadlock is actually reachable (but not necessarily minimal), by filling each channel with a header flit.\n");
	}
	else if (dl == RET_DLF) {
		if (output)
    			std::cout << "The network is deadlock-free under configuration with faulty channels: " << *faulty_resources << std::endl;
	}
	else if (dl == RET_DISC_ROUTING) {
		if (output)
			printf("The network has a disconnected routing function.\n");
	}
	else
		assert(false);

TERMINATE_RUN_ALGORITHM:
    //write(fildes, &dl, sizeof(dl));

    for (i=0; i<num_of_resources; i++) {
        if (nodeset[i] != 0) {
            for (j=0; j<g[i].n; j++)
                free(g[i].labels(j));
            free(g[i].parents);
        }
    }
    delete[] g;
    delete[] nodeset;
    delete[] marks;
    delete[] visited_013;
if (dl != RET_DLF && !ASSEMBLE_STATISTICS) exit(0);
    return dl;
}

int num_of_cores() {
#ifdef __APPLE__
    int mib[4], numCPU;
    size_t len = sizeof(numCPU);

    /* set the mib for hw.ncpu */
    mib[0] = CTL_HW;
    mib[1] = HW_AVAILCPU;  // alternatively, try HW_NCPU;

    /* get the number of CPUs from the system */
    sysctl(mib, 2, &numCPU, &len, NULL, 0);

    if( numCPU < 1 ) {
        mib[1] = HW_NCPU;
        sysctl( mib, 2, &numCPU, &len, NULL, 0 );

        if( numCPU < 1 ) {
            numCPU = 1;
        }
    }
    return numCPU;
#else
    return sysconf( _SC_NPROCESSORS_ONLN );
#endif
}
