#include <zmq.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/wait.h>
#include <sstream>

#include "general.h"

void freeZMQ(void *data, void *hint) {
    free(data);
}

void clientLoop(const char *host, Params *params) {
    void *context = zmq_init (1);

    //  Socket to talk to server
    printf ("Connecting to server...\n");
    void *requester = zmq_socket (context, ZMQ_REQ);
    zmq_connect (requester, host);

    char stop = 0;

    //Resource*** list_of_frs = compute_faulty_resources(params);

	std::cout << "INIT" << std::endl;
        zmq_msg_t request;
        zmq_msg_init_size (&request, 5);
        memcpy (zmq_msg_data (&request), "INIT", 4);
        zmq_send (requester, &request, 0);
        zmq_msg_close (&request);

	std::cout << "waiting on reply to INIT" << std::endl;
        zmq_msg_t reply;
        zmq_msg_init (&reply);
        zmq_recv (requester, &reply, 0);
        zmq_msg_close (&reply);


    while (!stop) {
	std::cout << "REQ" << std::endl;
        zmq_msg_t request;
        zmq_msg_init_size (&request, 5);
        memcpy (zmq_msg_data (&request), "REQ", 3);
        //printf ("Requesting work...\n");
        zmq_send (requester, &request, 0);
        zmq_msg_close (&request);

        zmq_msg_t reply;
        zmq_msg_init (&reply);
        zmq_recv (requester, &reply, 0);
	const char* cmd = (const char*)zmq_msg_data(&reply);
	std::stringstream ss(cmd);
	int n, batch;
	ss >> n >> batch;
        //printf ("Received number %d\n", n);
        zmq_msg_close (&reply);

        int error = run_algorithm_on_batch_of_faulty_configurations(n, n+batch, params);
        if (error) {
            //stop = 1;
            char buffer[1024];
            int len = snprintf(buffer, sizeof(buffer), "DEADLOCK found in config %i\n", n);
            zmq_msg_init_data(&request, strdup(buffer), len+1, freeZMQ, NULL);
            zmq_send (requester, &request, 0);
            zmq_msg_close (&request);
        } else {
		std::cout << "ACK" << std::endl;
		char buffer[128];
		int size = snprintf(buffer, sizeof(buffer)-1, "ACK %s", cmd);
        	zmq_msg_t request;
        	zmq_msg_init_size (&request, size+1);
        	memcpy (zmq_msg_data (&request), buffer, size+1);
        	zmq_send (requester, &request, 0);
        	zmq_msg_close (&request);
	}

	// reply on ACK/DEADLOCK found
        zmq_msg_init (&reply);
        zmq_recv (requester, &reply, 0);
        zmq_msg_close (&reply);
    }
    zmq_close (requester);
    zmq_term (context);
}

int main (int argc, char **argv)
{
    if (argc <= 1) {
        fprintf(stderr, "usage: %s [hostname]\n", argv[0]);
        return -1;
    }
    char *hostname = argv[1];
    char host[256];
    snprintf(host, sizeof(host), "tcp://%s:5555", hostname);
    fprintf(stdout, "Server: %s\n", host);

    double load[3];
    getloadavg(load, 3);

    Params *params = inst_params();

    int cores = num_of_cores();
    int childs = cores-load[0]+1;

    fprintf(stderr, "starting %i childs\n", childs);

    if (childs <= 0)
        return 0;

    int i;
    for (i=0; i<childs; ++i) {
        pid_t pid = fork();
        if (pid == 0) {
            clientLoop(host, params);
            _exit(0);
        }
    }
    while (childs > 0) {
        int status;
        int pid = wait(&status);
        if (pid > 0)
            childs--;
    }
    return 0;
}
