#ifndef DATASTRUCTURES
#define DATASTRUCTURES

#include <set>
#include <vector>
#include <iostream>
#include <assert.h>

const int RET_DLF = 0;
const int RET_DL_MIN = 1;
const int RET_DL_NOT_MIN = 2;
const int RET_ERROR = -1;
const int RET_DISC_ROUTING = -2;
const int RET_OK = -3;

/*
DATA STRUCTURE:
Vertices are integers, from 0 to |V|.
A graph is a list of lists. The nth list represents the set of neighbors of the nth node. Thus, say g is a graph structure, then:
g[v].n returns the number of neighbors;
g[v].edges[n].neighbor returns the nth neighbor of v;
g[v].edges[n].labels returns the labels of edge (v, v_1), where v_1 is the nth neighbor of v.
*/
//#define NEIGHBORS_IN_EDGE

struct Edge {
    int *labels;
	bool labels_added_to_imm;
#ifdef NEIGHBORS_IN_EDGE
    unsigned short neighbor;
#endif
    unsigned char num_of_labels;
};

class Graph {
    struct Edge edges[NUM_OF_OUTGOING_EDGES]; // list of neighbors with labels
#ifndef NEIGHBORS_IN_EDGE
    unsigned short neighbors[NUM_OF_OUTGOING_EDGES];
#endif
    public:
    int n;		// num of neighbors

    int *parents; // list of parents 
    int parents_allocated;	// the number of allocated memory for parents
    int parents_elements;	// the number of elements in parents

    int neighbor(int i) const {
#ifdef NEIGHBORS_IN_EDGE
        return edges[i].neighbor;
#else
        return neighbors[i];
#endif
    }
    bool labelsAddedToImm(int i) {
        return edges[i].labels_added_to_imm;
    }
    void setLabelsAddedToImm(int i, bool v) {
        edges[i].labels_added_to_imm = v;
    }
    int *labels(int i) {
        return edges[i].labels;
    }
    int labelCount(int i) {
        return edges[i].num_of_labels;
    }
    void setLabelCount(int i, int v) {
        edges[i].num_of_labels = v;
    }
    void set(int i, bool labels_added_to_imm, int neighbor, int num_of_labels, int *labels) {
        assert(i >= 0 && i < NUM_OF_OUTGOING_EDGES);
#ifdef NEIGHBORS_IN_EDGE
        edges[i] = (Edge){labels, labels_added_to_imm, neighbor, num_of_labels};
#else
        edges[i] = (Edge){labels, labels_added_to_imm, (unsigned char)num_of_labels};
        neighbors[i] = neighbor;
#endif
        //std::cout << num_of_labels << std::endl;
    }
};

/*
   struct Marks {
   int marking;	// 0,1,2,3,4
   int *sens;		// deadlock-sensitive destinations
   int *imm;		// deadlock-immune destinations

   int sens_allocated; // the number of allocated memory for sens
   int sens_elements;	// the number of elements in sens
   int imm_allocated;  // the number of allocated memory for imm
   int imm_elements;	// the number of elements in imm
   };
   */
struct Marks {
    int marking;	// 0,1,2,3,4
    typedef std::vector<int> Container;
    Container sens;     // deadlock-sensitive destinations
    Container imm;      // deadlock-immune destinations

    int subsetNumber; // first 'subsetNumber' elements of sens are in imm
    int immPlace;
    Marks() : marking(0), sens(), imm(), subsetNumber(0), immPlace(-1) {
    }
    bool subset();
};


struct int_List {
    short n;
    int *elts;
};

struct dyn_int_List {
    short num_elts;
    short num_allocated;
    int *elts;
};

//int subsetp(int *a, int *b, int len_a, int len_b);

void init_dyn_int_List(struct dyn_int_List *x, int n);

/* Structure storing a constraint
        pos[0]lhs[0] + pos[0]lhs[1] + ... pos[lhs_size-1]lhs[lhs_size-1] x rhs
        where:
                x is determined by the constr_type (LE, EQ, GE, OBJ)
                lhs[n] >= 0 denotes a variable
                lhs[n] < 0 denotes the constant non-negative integer value -(lhs[n]+1)
*/
struct constraint {
        int* lhs;
        int* pos;
        int lhs_size;
        int rhs;
        int constr_type;
};
const int LE = 1;
const int GE = 2;
const int EQ = 3;
const int OBJ = 4;

void print_constraint(struct constraint c);


struct RET_table {
	int dl;
	int dlf;
	int disc_routing;
	bool terminate;
};


void add_ret_tables(struct RET_table * t1, struct RET_table t2);
void add_result_to_ret_table(struct RET_table * t, int result);

#endif
