/*********************/
/* 2D TORUS ROUTING */
/********************/
#include<math.h>
#include "2d_torus_vc_definitions.h"
int ksize_x=PARAM_x;
int ksize_y=PARAM_y;

int checkWraplink(int xoff,int yoff){
	int cnt=0;
	if((fabs(xoff)>(double)ksize_x/2)||(fabs(yoff)>(double)ksize_y/2))
		cnt=1;
	return cnt;
}

/* The routing function. Routes from channel c == (x,y,p) to processing node d == (dx,dy)
   Returns a set of resources. 
   This list of one element is closed with a NULL.
*/

ResourceList* inst_routing(const Resource &c, Procnode* d, const Params* dim, ResourceList* frs) {
    ResourceList *hops = new ResourceListN<7>;
    Resource *nexthops = hops->routed;
	int i =0;


	// The coordinates of the destination
	int dx = d->x, dy = d->y;
	// The coordinates of the processing node at the end of channel c
	Procnode end = get_end(c);
	int nx = end.x, ny = end.y;
	// The offset of the processing node and the desination node
	int xoff=dx-nx, yoff=dy-ny;
	if (xoff != 0 || yoff != 0 ) {										// If we do not arrive at our destination
		
		// Case1. The message do need to traverse the wraparound link	
		if (checkWraplink(xoff, yoff)){
			if (((xoff>0&&xoff<=(double)ksize_x/2)||xoff<(double)-ksize_x/2)){				
			//if (nx != ksize_x-1 && ((xoff>0&&xoff<=(double)ksize_x/2)||xoff<(double)-ksize_x/2)){				
				nexthops[i++] = Resource(nx,ny,E1);
			}
			if (((xoff<0&&xoff>=(double)-ksize_x/2)||xoff>(double)ksize_x/2)) {
			//if (nx!=0 && ((xoff<0&&xoff>=(double)-ksize_x/2)||xoff>(double)ksize_x/2)) {
				nexthops[i++] = Resource(nx,ny,W1);
			}
			if (((yoff>0&&yoff<=(double)ksize_y/2)||yoff<(double)-ksize_y/2)){
			//if (ny != ksize_y-1 && ((yoff>0&&yoff<=(double)ksize_y/2)||yoff<(double)-ksize_y/2)){
				nexthops[i++] = Resource(nx,ny,N1);
			}
			if (((yoff<0&&yoff>=(double)-ksize_y/2)||yoff>(double)ksize_y/2)) {
			//if (ny!=0 && ((yoff<0&&yoff>=(double)-ksize_y/2)||yoff>(double)ksize_y/2)) {
				nexthops[i++] = Resource(nx,ny,S1);

/*
With the first cases (VCs in the wraparound links), a deadlock occurs for 7x7 and greater meshes only (4 worms).
With the second cases (no VCs in the wraparound links), a 6-worm deadlock occurs in a 5x5 mesh as well.
*/
			}

			//Provide the wraparound linke when the node is the bound node in the lowest dimension(need to pass wraparound links)
			if(xoff > ksize_x/2.0 && nx == 0){
				nexthops[i++] = Resource(nx,ny,W2);
			}
			if(xoff < -ksize_x/2.0 && nx == ksize_x-1){
				nexthops[i++] = Resource(nx,ny,E2);
			}
			if(fabs(xoff) <= ksize_x/2.0){
				if(yoff > ksize_y/2.0&& ny == 0){
					nexthops[i++] = Resource(nx,ny,S2);
				}
				if(yoff < -ksize_y/2.0&& ny == ksize_y-1){
					nexthops[i++] = Resource(nx,ny,N2);
				}
			}	
assert(i>0);
		}
		
		// Case2. The message do not need to traverse the wraparound link
		else{
			//VC1 provide a fully adaptive routing for shortpath routing
			if(xoff < 0)
				nexthops[i++] = Resource(nx,ny,W1);
			if(xoff > 0)
				nexthops[i++] = Resource(nx,ny,E1);
			if(yoff < 0)
				nexthops[i++] = Resource(nx,ny,S1);
			if(yoff > 0)
				nexthops[i++] = Resource(nx,ny,N1);
				
			//VC2 provide a dimension order routing for shortpath routing				
			if(xoff != 0){
				if(xoff < 0)
					nexthops[i++] = Resource(nx,ny,W2);
				if(xoff > 0)				
					nexthops[i++] = Resource(nx,ny,E2);
			}
			else{
				if(yoff < 0)
					nexthops[i++] = Resource(nx,ny,S2);
				if(yoff > 0)
					nexthops[i++] = Resource(nx,ny,N2);
			}
		}
	}
	return hops;
}

