/**************/
/* XYZ ROUTING */
/**************/
#include "3d_mesh_definitions.h"

/* The routing function. Routes from channel c == (x,y,z,p) to processing node d == (dx,dy,dz)
   Returns a set of resources. As xyz routing is deterministic, only one next hop is returned.
   This list of one element is closed with a NULL.
*/
ResourceList* inst_routing(const Resource &c, Procnode* d, const Params* dim, ResourceList* frs) {
    ResourceList *hops = new ResourceListN<2>;
    Resource *nexthops = hops->routed;

	// The coordinates of the destination
	int dx = d->x, dy = d->y, dz = d->z;
	// The coordinates of the processing node at the end of channel c
	int nx = get_end(c).x, ny = get_end(c).y, nz = get_end(c).z;

	if (dx != nx || dy != ny || dz != nz) {				// If we do not arrive at our destination
		if (dx < nx)
			nexthops[0] = Resource(nx,ny,nz,W);	//go west if the destination is west
		else if (dx > nx)
			nexthops[0] = Resource(nx,ny,nz,E); 	//go east if the destination is east
		else if (dy < ny)
			nexthops[0] = Resource(nx,ny,nz,S);	//go south if the destination is southh && in the right column
		else if (dy > ny)
			nexthops[0] = Resource(nx,ny,nz,N);	//go north if the destination is north && in the right column
		else if (dz > nz)
			nexthops[0] = Resource(nx,ny,nz,U);	//go up if the destination is up && in the right column && in the right row
		else if (dz < nz)
			nexthops[0] = Resource(nx,ny,nz,D);	//go down if the destination is down && in the right column && in the right row
	}
	return hops;
}
