/*************************/
/* 3D XYZ  TORUS ROUTING */
/*************************/
#include    <math.h>
#include "3d_torus_vc_definitions.h"
#define  SIZE  8;    					// The size of the X/Y/Z dimension
int ksize=SIZE;

/*	Check if the message has passed the x/y/z wraparound link, if the message has not passed
	the x wraparound link, then the vc2 channels in dimension x will be used, while the vc1
	channels in dimension x will be used.
	The same apply to dimension y and z.
*/


void decide_vc_number(const Resource &c, int VCID[]){
	Procnode src = get_src(c), end = get_end(c);
	
	int x1 = src.x, y1 = src.y, z1 = src.z;
	int x2 = end.x, y2 = end.y, z2 = end.z;
	
	if(abs(x1-x2)==ksize-1 || c.p == W2 || c.p == E2){
		VCID[0] = 2;
	}
	if(abs(y1-y2)==ksize-1 || c.p == N2 || c.p == S2){
		VCID[1] = 2;
	}
	if(abs(z1-z2)==ksize-1 || c.p == U2 || c.p == D2){
		VCID[2] = 2;
	}	
}


/* The routing function. Routes from channel c == (x,y,z,p) to processing node d == (dx,dy,dz)
   Returns a set of resources. This list of one element is closed with a NULL.
*/

ResourceList* inst_routing(const Resource &c, Procnode* d, const Params* dim, ResourceList* frs) {
    ResourceList *hops = new ResourceListN<10>;
    Resource *nexthops = hops->routed;


	int i = 0;
	// The coordinates of the destination
	int dx = d->x, dy = d->y, dz = d->z;
	// The coordinates of the processing node at the end of channel c
	Procnode end = get_end(c);
	int nx = end.x, ny = end.y, nz = end.z;
	// The offset of the desination node  and the processing node
	int xoff=dx-nx, yoff=dy-ny, zoff=dz-nz;
	int VCID[3]={1,1,1};
	
	if (xoff != 0 || yoff != 0 || zoff != 0) {										// If we do not arrive at our destination
		decide_vc_number(c,VCID);														//Decide the virtual channel number 1/2 to be used
		//dimension order algorithm
		if(xoff != 0){																//if xoffset is not 0
			if ((xoff > 0&&xoff <=ksize/2)||xoff < -ksize/2.0){
				if(VCID[0]==1){
					nexthops[i++] = Resource(nx,ny,nz,E1);
				}
				if(VCID[0]==2){
					nexthops[i++] = Resource(nx,ny,nz,E2);
				}
			}
			if ((xoff < 0&&xoff >= -ksize/2.0)||xoff > ksize/2.0){
				if(VCID[0]==1){
					nexthops[i++] = Resource(nx,ny,nz,W1);
				}
				if(VCID[0]==2){
					nexthops[i++] = Resource(nx,ny,nz,W2);
				}
			}
		}
		else{																		//if xoffset is 0
			if(yoff!=0){															//if yoffset is not 0
				if ((yoff>0&&yoff <=ksize/2.0)||yoff < -ksize/2.0){
					if(VCID[1]==1){
						nexthops[i++] = Resource(nx,ny,nz,N1);
					}
					if(VCID[1]==2){
						nexthops[i++] = Resource(nx,ny,nz,N2);
					}
				}
				if ((yoff < 0&&yoff>=-ksize/2.0)||yoff > ksize/2.0){
					if(VCID[1]==1){
						nexthops[i++] = Resource(nx,ny,nz,S1);
					}
					if(VCID[1]==2){
						nexthops[i++] = Resource(nx,ny,nz,S2);
					}
				}
			}
			else {																		//if zoffset is 0
				if ((zoff>0&&zoff <=ksize/2.0)||zoff < -ksize/2.0){
					if(VCID[2]==1){
						nexthops[i++] = Resource(nx,ny,nz,U1);
					}
					if(VCID[2]==2){
						nexthops[i++] = Resource(nx,ny,nz,U2);
					}
				}
				if ((zoff < 0&&zoff>=-ksize/2.0)||zoff > ksize/2.0){
					if(VCID[2]==1){
						nexthops[i++] = Resource(nx,ny,nz,D1);
					}
					if(VCID[2]==2){
						nexthops[i++] = Resource(nx,ny,nz,D2);
					}
				}
			}
		}
			
	}
	return hops;
}


