/**************/
/* ROUTING    */
/**************/

/*
Routing function for Diagional Mesh NoC from UCI.
*/

#include "diagonal_definitions.h"

// The default NePA routing function: adaptively choose a shortest path towards the destination
// Subnetwork 1 has eastern going traffic, subnetwork 2 western.
ResourceList* inst_routing(const Resource &c, Procnode* d, const Params* dim, ResourceList* frs) {
    ResourceList *hops = new ResourceListN<4>;
    Resource *nexthops = hops->routed;
	int i = 0;

	// The coordinates of the processing node at the end of channel c
	int nx = get_end(c).x, ny = get_end(c).y;
	// The coordinates of the destination
	int dx = d->x, dy = d->y;
	// Adaptive routing in NePA architecture
	if (c.p == S1 || c.p == N1 || c.p == E || c.p == SE || c.p == NE) {
		// SUBNETWORK 1: going in eastern direction
		if (dx > nx)
			nexthops[i++] = Resource(nx, ny, E);
		if (dy > ny)
			nexthops[i++] = Resource(nx, ny, S1);
		if (dy < ny)
			nexthops[i++] = Resource(nx, ny, N1);
		if (dx > nx && dy > ny)
			nexthops[i++] = Resource(nx, ny, SE);
		if (dx > nx && dy < ny)
			nexthops[i++] = Resource(nx, ny, NE);
	}
	if (c.p == S2 || c.p == N2 || c.p == W || c.p == SW || c.p == NW) {
		// SUBNETWORK 2: going in western direction
		if (dx < nx)
			nexthops[i++] = Resource(nx, ny, W);
		if (dy > ny)
			nexthops[i++] = Resource(nx, ny, S2);
		if (dy < ny)
			nexthops[i++] = Resource(nx, ny, N2);
		if (dx < nx && dy > ny)
			nexthops[i++] = Resource(nx, ny, SW);
		if (dx < nx && dy < ny)
			nexthops[i++] = Resource(nx, ny, NW);
	}
	if (c.p == L) {
		// LOCAL: packet needs to be injected into a subnetwork
		if (dx > nx) {
			// Choose Subnetwork 1
			nexthops[i++] = Resource(nx, ny, E);
			if (dy > ny) {
				nexthops[i++] = Resource(nx, ny, S1);
				nexthops[i++] = Resource(nx, ny, SE);
			}
			if (dy < ny) {
				nexthops[i++] = Resource(nx, ny, N1);
				nexthops[i++] = Resource(nx, ny, NE);
			}
		}
		if (dx < nx) {
			// Choose Subnetwork 2
			nexthops[i++] = Resource(nx, ny, W);
			if (dy > ny) {
				nexthops[i++] = Resource(nx, ny, S2);
				nexthops[i++] = Resource(nx, ny, SW);
			}
			if (dy < ny) {
				nexthops[i++] = Resource(nx, ny, N2);
				nexthops[i++] = Resource(nx, ny, NW);
			}
		}
		if (dx == nx) {
			// Arbitrarily, choose a subnetwork
			if (dy > ny) {
				nexthops[i++] = Resource(nx, ny, S1);
				nexthops[i++] = Resource(nx, ny, S2);
			}
			if (dy < ny) {
				nexthops[i++] = Resource(nx, ny, N1);
				nexthops[i++] = Resource(nx, ny, N2);
			}
		}
	}
	return hops;
}



