/**
 * TODO:
 * - keep track of jobs, and only stop if all jobs are finished... (now nobody waits on the last jobs)
 * - client should request after each iteration if it should execute another, by request the load (and substracting your own influence)
 */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <sys/sysctl.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <strings.h>
#include <sstream>

#include "general.h"

#include <zmq.h>
#include <unistd.h>

void freePayload(void *data, void *hint) {
	free(data);
}

std::set<int> acked;

int main(int argc, char **argv) {
	int batchSize = 100;
	if (argc > 1) {
		batchSize = atoi(argv[1]);
	}
	// The starting time, for bookkeeping
	time_t start;
	time(&start);
	// Instantiate the topology parameters, i.e., the size of a mesh
	Params *params = inst_params();
	// Compute the number of different configurations to test
	int num_configs = get_num_of_faulty_configurations(params);
	int current = 0;
	// Prepare our context and socket
	void *context = zmq_init (1);
	int faults = 0;
	int workers = 0;

	// Socket to talk to clients
	void *responder = zmq_socket (context, ZMQ_REP);
	zmq_bind (responder, "tcp://*:5555");

	fprintf(stderr, "# configs = %i\n", num_configs);
	int i = 0;

	while (acked.size() < num_configs/100) {
		zmq_msg_t request;
		zmq_msg_init (&request);
		zmq_recv (responder, &request, 0);

		char isAck = memcmp(zmq_msg_data(&request),"ACK", 3) == 0;
		char isRequest = memcmp(zmq_msg_data(&request),"REQ", 3) == 0;
		char isInit = memcmp(zmq_msg_data(&request),"INIT", 4) == 0;

		if (isInit) {
			workers++;

			zmq_msg_t reply;
			zmq_msg_init_data(&reply, strdup(""), 1, freePayload, NULL);
			zmq_send (responder, &reply, 0);
			zmq_msg_close (&reply);

		} else if (isRequest) {
			// at least one block
			do {
				current = (current + batchSize) % num_configs;
			} while (acked.find(current) != acked.end());

			char buffer[64];
			int len = snprintf(buffer, sizeof(buffer)-1, "%i %i", current, batchSize);

			// Send reply back to client
			char *payload = strdup(buffer);

			zmq_msg_t reply;
			zmq_msg_init_data(&reply, payload, len+1, freePayload, NULL);
			zmq_send (responder, &reply, 0);
			zmq_msg_close (&reply);
		} else if (isAck) {
			fprintf(stdout, ".");
			fflush(stdout);
			if (++i % 100 == 0)
				fprintf(stdout, " -> %i / %i (%f%% using %i nodes)\n", current, num_configs, (float)current*100/num_configs, workers);

			const char* cmd = (const char*)zmq_msg_data(&request);
			std::stringstream ss(&cmd[4]);
			int n, batch;
			ss >> n >> batch;
			//printf ("Received acked for '%s' -> %d:%d\n", cmd, n, batch);
			acked.insert(n);

			zmq_msg_t reply;
			zmq_msg_init_data(&reply, strdup(""), 1, freePayload, NULL);
			zmq_send (responder, &reply, 0);
			zmq_msg_close (&reply);
		} else {
			fwrite(zmq_msg_data(&request), zmq_msg_size(&request), 1, stderr);
			faults++;
			//abort();

			zmq_msg_t reply;
			zmq_msg_init_data(&reply, strdup(""), 1, freePayload, NULL);
			zmq_send (responder, &reply, 0);
			zmq_msg_close (&reply);
		}

		zmq_msg_close (&request);
	}

	time_t end;
	time(&end);
	printf("\n\n");
	printf("Number of cores used: %i\n", workers);
	printf("Number of faults: %i\n", faults);
	printf("Starting time: %s", ctime(&start));
	printf("End time: %s", ctime(&end));
	printf("Elapsed time: %lli seconds\n", (long long int)(end-start));

	return 0;
}
