#ifndef _STACK_H_
#define _STACK_H_

template <class T, T *T::*ptr = &T::next>
class Stack {
    T *value;
public:
    inline Stack(T *start = NULL) :
        value(start) {
    }
    inline Stack(Stack<T> &c) {
        value = c.value;
        c.value = NULL;
    }
    virtual inline ~Stack() {
        clear();
    }

    void clear() {
        value = NULL;
    }
    inline void put(T *e) {
        if (e) {
            e->*ptr = value;
            value = e;
        }
    }

    inline bool put(Stack &c) {
        if (&c == this || c.empty())
            return false;
        if (!empty()) {
            T *last = c.value;
            while (last && last->*ptr) {
                last = last->*ptr;
            }
            if (last)
                last->*ptr = value;
        }
        value = c.value;
        return true;
    }

    inline T *get() {
        T *e = value;
        if (e) {
            value = (T*)e->*ptr;
            e->*ptr = NULL;
        }
        return e;
    }

    inline Stack getAll() {
        Stack retval;
        retval.value = value;
        value = NULL;
        return retval;
    }

    /*
    template <void (T::*func)()>
    inline void forEach() const {
        forEach< &invoke<T,func> >();
    }
    */

    template <void (*func)(T*)>
    inline void forEach() const {
        T *current = value;
        while (current) {
            T *next = (T*)current->*ptr;
            func(current);
            current = next;
        }
    }

    inline void forEach(void (*func)(T*)) const {
        T *current = value;
        while (current) {
            T *next = (T*)current->*ptr;
            func(current);
            current = next;
        }
    }

    inline bool empty() const {
        return !value;
    }

    class iterator {
        T *value;
    public:
        iterator(T *_value) : value(_value) {
        }
        ~iterator() {
        }
        inline T *operator*() {
            return value;
        }
        inline T *operator->() {
            return &*value;
        }
        inline iterator &operator++() {
            value = (T*)value->*ptr;
            return *this;
        }
        inline bool operator!=(iterator b) {
            return b.value != value;
        }
        inline bool operator==(const iterator &b) {
            return b.value == value;
        }
        inline operator bool() {
            return value;
        }
    };

    inline iterator begin() {
        return iterator(value);
    }
    inline iterator end() {
        return iterator(NULL);
    }
};    

#endif


